import numpy as np
from sklearn.datasets import load_svmlight_file
from numpy import linalg as LA

class Worker(object):
    def __init__(self, rank, n_components,
                 kernel='rbf', gamma=0.1,
                 weight_seed=0, bias_seed=0):
        self.rank = rank
        self.n_components = n_components
        self.kernel = kernel
        self.gamma = gamma
        self.weight_seed = weight_seed
        self.bias_seed = bias_seed
        self.weight_generator = np.random.RandomState(self.weight_seed)
        self.bias_generator = np.random.RandomState(self.bias_seed)

    # def randnum_gen(self):
    #     self.generator = np.random.RandomState(self.random_state)

    def rand_fourier(self, x):
        """ Generate random Fourier features """
        n_feature = x.shape[1]

        random_weight = np.sqrt(2 * self.gamma) * self.weight_generator.normal(size=(n_feature, self.n_components))
        bias = self.bias_generator.uniform(0, 2*np.pi, self.n_components)

        projection = np.matmul(x, random_weight) + bias

        cos_feature = np.cos(projection)

        random_feature = np.sqrt(2. / self.n_components) * cos_feature

        return random_feature

    def construction(self, random_feature, eta, lamb=None, local_eigvecs=None, eigvals=None):
        if eigvals is None:
            local_feature = np.sqrt(eta) * np.copy(random_feature)
        else:
            Z_root = np.matmul(local_eigvecs, np.diag(np.sqrt(eigvals-eta*lamb)))
            local_feature = np.concatenate((np.sqrt(1.0-eta) * Z_root, np.sqrt(eta) * random_feature), axis=1)

        return local_feature

    def local_computation(self, local_feature, vec):
        """ The local computation of decentralized power method """
        local_eig_vec = np.matmul(local_feature, vec)

        local_update_vec = np.matmul(local_feature.T, local_eig_vec)

        return local_update_vec, local_eig_vec




